import importlib
from PySide6.QtCore import Qt, QTimer
from PySide6.QtWidgets import (
    QGroupBox,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QPushButton,
)
from icarus_v2.gui.image_button import ImageButton
from time import sleep


# Control panel for manual and console operation
class DeviceControlPanel(QGroupBox):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.pulse_generator = None

        # Initialize buttons
        with importlib.resources.path('icarus_v2.resources', 'shutdown.svg') as shutdown_path:
            self.shutdown_button = ImageButton(str(shutdown_path), "Shutdown", cooldown_ms=100)
        self.shutdown_button.clicked.connect(self.on_shutdown)

        self.mode_button = QPushButton("Manual")
        self.mode_button.setObjectName("noEnabled")
        self.mode_button.setCheckable(True)
        self.mode_button.toggled.connect(self.toggle_mode)

        self.pump_button = QPushButton("Pump")
        self.pump_button.setObjectName("toggleButton")
        self.pump_button.setCheckable(True)
        self.pump_button.toggled.connect(self.toggle_pump)

        self.pressurize_button = QPushButton("Pressurize")
        self.pressurize_button.setObjectName("toggleButton")
        self.pressurize_button.setCheckable(True)
        self.pressurize_button.toggled.connect(self.toggle_pressurize)

        self.depressurize_button = QPushButton("Depressurize")
        self.depressurize_button.setObjectName("toggleButton")
        self.depressurize_button.setCheckable(True)
        self.depressurize_button.toggled.connect(self.toggle_depressurize)

        self.pulse_button = QPushButton("Pulsing")
        self.pulse_button.setObjectName("toggleButton")
        self.pulse_button.setCheckable(True)
        self.pulse_button.toggled.connect(self.toggle_pulsing)

        self.shutdown_button.setFixedSize(150, 150)
        policy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.mode_button.setSizePolicy(policy)
        self.pump_button.setSizePolicy(policy)
        self.pressurize_button.setSizePolicy(policy)
        self.depressurize_button.setSizePolicy(policy)
        self.pulse_button.setSizePolicy(policy)
        self.spacer_item = QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout = QVBoxLayout(self)
        self.layout.addWidget(self.shutdown_button, alignment=Qt.AlignHCenter | Qt.AlignTop)
        self.layout.addWidget(self.mode_button)
        self.layout.addWidget(self.pump_button)
        self.layout.addWidget(self.pressurize_button)
        self.layout.addWidget(self.depressurize_button)
        self.layout.addWidget(self.pulse_button)

        # Width of depressurize button.Without this the panel shrinks when buttons are hidden
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setStyleSheet("font-size: 21pt;")

    def toggle_mode(self, checked):
        if checked:
            self.mode_button.setText("Console")
            # Ensure the mode button doesn't expand or move
            size = self.mode_button.size().height()
            self.mode_button.setMinimumHeight(size)
            self.mode_button.setMaximumHeight(size)

            for button in [self.pump_button, self.pressurize_button, self.depressurize_button, self.pulse_button]:
                button.setVisible(False)
                button.setEnabled(False)

            self.layout.addItem(self.spacer_item)

            # Set pulser off, pump on, valves closed
            self.pulse_button.setChecked(False)
            self.pump_button.setChecked(True)
            self.pressurize_button.setChecked(False)
            self.depressurize_button.setChecked(False)
        else:
            self.mode_button.setText("Manual")
            for button in [self.pump_button, self.pressurize_button, self.depressurize_button, self.pulse_button]:
                button.setVisible(True)
                button.setEnabled(True)

            self.layout.removeItem(self.spacer_item)

            # Remove size constraint on mode button
            self.mode_button.setMaximumHeight(16777215)
            self.mode_button.setMinimumHeight(0)

    def toggle_pump(self, checked):
        if checked:
            self.pulse_generator.set_pump_low()
        else:
            self.pulse_generator.set_pump_high()

    def toggle_depressurize(self, checked):
        if checked:
            self.pulse_generator.set_depressurize_low()
        else:
            self.pulse_generator.set_depressurize_high()

    def toggle_pressurize(self, checked):
        if checked:
            self.pulse_generator.set_pressurize_low()
        else:
            self.pulse_generator.set_pressurize_high()

    def toggle_pulsing(self, checked):
        if checked:
            # Prevent other buttons from doing anything while pulsing
            self.pressurize_button.setEnabled(False)
            self.depressurize_button.setEnabled(False)

            # Close valves
            self.pressurize_button.setChecked(False)
            self.depressurize_button.setChecked(False)
            # Prevents 2 depressurize events in the same chunk
            # One chunk is update_rate^-1 = 30^-1 = 0.0333 seconds
            sleep(0.05)

            self.pulse_generator.start()
        else:
            # make sure no buttons are clicked until pulsing is fully done
            self.pulse_generator.quit()
            self.pulse_generator.wait()
            self.pressurize_button.setEnabled(True)
            self.depressurize_button.setEnabled(True)

    # Connect buttons to the pulse generator
    def set_pulse_generator(self, pulse_generator):
        self.pulse_generator = pulse_generator

    def on_shutdown(self):
        # turn pump off
        self.pump_button.setChecked(False)
        # stop pulsing
        self.pulse_button.setChecked(False)
        # open valves
        self.pressurize_button.setChecked(True)
        self.depressurize_button.setChecked(True)

        # wait 3 seconds and then set the buttons to unchecked
        # only if currently in manual, as in console mode closing the valves returns control to console
        if not self.mode_button.isChecked():
            self.pressurize_button.setEnabled(False)
            self.depressurize_button.setEnabled(False)
            self.pump_button.setEnabled(False)
            self.pulse_button.setEnabled(False)
            QTimer.singleShot(1000, self.reset_valves)

        # revert to manual mode
        self.mode_button.setChecked(False)

    def reset_valves(self):
        self.pressurize_button.setEnabled(True)
        self.depressurize_button.setEnabled(True)
        self.pump_button.setEnabled(True)
        self.pulse_button.setEnabled(True)

        self.pressurize_button.setChecked(False)
        self.depressurize_button.setChecked(False)

    def reset(self):
        # Reset buttons to initial states without triggering their slots
        for button in [self.pump_button, self.pressurize_button, self.depressurize_button, self.pulse_button]:
            button.blockSignals(True)
            button.setChecked(False)
            button.blockSignals(False)

        # Change to manual mode
        self.mode_button.setChecked(False)



import os
from pyqtgraph import PlotWidget
from icarus_v2.backend.configuration_manager import ConfigurationManager
from icarus_v2.backend.event import Channel, HistStat, Event
from icarus_v2.qdarktheme.load_style import THEME_COLOR_VALUES, ACCENT_COLORS, url, color
import pyqtgraph as pg
from pyqtgraph.graphicsItems.ButtonItem import ButtonItem
from PySide6.QtCore import QEvent, QStandardPaths
from PySide6.QtGui import Qt
from PySide6.QtWidgets import QDialog, QPushButton, QVBoxLayout, QFileDialog, QLabel, QSizePolicy
from icarus_v2.backend.csv_exporter import CSVExporter
import numpy as np
from bisect import bisect_left, bisect_right


class StyledPlotWidget(PlotWidget):
    def __init__(self, x_zoom=False):
        self.full_init=False

        self.config_manager = ConfigurationManager()
        self.config_manager.settings_updated.connect(self.update_theme)

        background = self.get_background_color()
        PlotWidget.__init__(self, background=background)

        self.showGrid(x=True, y=True, alpha=0.3)
        self.setMouseEnabled(x=x_zoom, y=False)  # Prevent zooming
        self.hideButtons()  # Remove autoScale button
        self.getPlotItem().getViewBox().setMenuEnabled(False) # Remove right click menu

        # Set up the export button
        self.export_button = ButtonItem( width=20, parentItem=self.plotItem)
        self.export_button.clicked.connect(self.export_clicked)
        self.export_button.setPos(0, 210)
        self.export_button.hide()
        self.update_export_icon()

        # Enable hover events. Needed so that button is only visible when hovering over graph
        self.setAttribute(Qt.WA_Hover)

        # Lines
        self.lines = {}
        self.line_visibility = {}

        # Export functionality
        self.edit_dialog = None

        # Mouse coordinates
        self.mouse_label = QLabel("")
        size = 14
        self.mouse_label.setStyleSheet(f"font-size: {size}px;")
        self.mouse_label.setFixedSize(200, 17)
        self.mouse_label.setAlignment(Qt.AlignRight)

        # Add statistics tracking
        self.statistics = {}
        self.stat_displays = {}
        self.stat_layout = QVBoxLayout()
        if x_zoom:
            self.getPlotItem().getViewBox().sigStateChanged.connect(self.update_statistics)


        layout = QVBoxLayout()
        layout.addLayout(self.stat_layout)
        layout.addStretch(1)
        layout.addWidget(self.mouse_label, alignment=Qt.AlignRight)
        layout.setContentsMargins(0, 35, 5, 45)

        self.setLayout(layout)

        self.full_init=True

    def update_theme(self, settings_key: str | None = None, size: int = 14) -> None:
        # skip setting theme if this is triggered by settings updates
        if settings_key is not None and settings_key != "theme":
            return

        background = self.get_background_color()
        text_color = self.get_text_color()

        self.setBackground(background)

        plot = self.getPlotItem()
        plot.titleLabel.setText(plot.titleLabel.text, color=text_color)
        pen = pg.mkPen(color=text_color)
        plot.getAxis('left').setPen(pen)
        plot.getAxis('bottom').setPen(pen)

        for channel, line in self.lines.items():
            style = self.get_line_style(channel)
            pen = pg.mkPen(color=style[0], style=style[1])
            line.setPen(pen)

        for channel, stats in self.statistics.items():
            for _, stat_info in stats.items():
                line_color = self.get_line_style(channel)[0]
                stat_info['label'].setStyleSheet(f"color: {line_color}; font-size: {size}px;")

        self.update_export_icon()

    def set_title(self, title):
        text_color = self.get_text_color()
        self.setTitle(title, color=text_color, size="17pt")

    def set_y_label(self, label):
        text_color = self.get_text_color()
        self.setLabel('left', label, **{'color': text_color, 'font-size': '8pt'})

    def set_x_label(self, label):
        text_color = self.get_text_color()
        self.setLabel('bottom', label, **{'color': text_color, 'font-size': '8pt'})

    # Add a new line to the plot with given style.
    def add_line(self, channel):
        style = self.get_line_style(channel)
        pen = pg.mkPen(color=style[0], style=style[1])
        self.lines[channel] = self.plot([], [], name=str(channel), pen=pen)
        self.line_visibility[channel] = True
        return self.lines[channel]

    # Update data for a specific line.
    def update_line_data(self, channel, x_data, y_data, srate=None):
        if channel in self.lines:
            self.lines[channel].setData(x_data, y_data)

            # Update statistics if they exist for this line
            if channel in self.statistics:
                for format_str, stat_info in self.statistics[channel].items():
                    if len(y_data) > 0:
                        if srate is None:
                            value = stat_info['method'](y_data)
                        else:
                            value = stat_info['method'](y_data, srate)
                        stat_info['label'].setText(format_str.format(value))
                    else:
                        stat_info['label'].setText(format_str.format(0))

    def append_points(self, points_dict, x_point):
        """
        Append points at the same x value.

        Args:
            points_dict: Dictionary mapping line channels to y values
            x_point: x value
        """
        for channel, y_point in points_dict.items():
            if channel in self.lines:
                line = self.lines[channel]
                # Get existing data
                x_data, y_data = line.getData()
                if x_data is None:
                    x_data = []
                    y_data = []

                # Append new point
                x_data = np.append(x_data, x_point)
                y_data = np.append(y_data, y_point)

                # Update the line
                line.setData(x_data, y_data)

    # Show/hide a specific line.
    def toggle_line_visibility(self, channel, visible):
        if channel in self.lines:
            if visible and not self.line_visibility[channel]:
                self.addItem(self.lines[channel])
            elif not visible and self.line_visibility[channel]:
                self.removeItem(self.lines[channel])
            self.line_visibility[channel] = visible

    def add_statistic(self, channel, method, format_str, size=14):
        """
        Add a statistic display for a specific line.

        Args:
            channel: The line channel this statistic is associated with
            method: Function that takes an array and returns the statistic value
            format_str: Format string for the value (e.g., "Avg: {:.3f}")
            size: Font size in pixels
        """
        if channel not in self.lines:
            raise KeyError(f"Line with channel '{channel}' does not exist")

        label = QLabel(format_str.format(0))
        line_color = self.get_line_style(channel)[0]
        label.setStyleSheet(f"color: {line_color}; font-size: {size}px;")

        # Add to layout
        label.setFixedSize(200, 18)
        label.setAlignment(Qt.AlignRight)
        self.stat_layout.addWidget(label, alignment=Qt.AlignRight)

        # Store the statistic info
        if channel not in self.statistics:
            self.statistics[channel] = {}
        self.statistics[channel][format_str] = {
            'method': method,
            'label': label,
        }

    def update_statistics(self):
        """
        Update all statistics based on currently visible data
        """
        for channel, stats in self.statistics.items():
            for format_str, stat_info in stats.items():
                if channel in self.lines:
                    line = self.lines[channel]
                    x_data, y_data = line.getData()

                    if x_data is None:
                        stat_info['label'].setText(format_str.format(0))
                        continue

                    view_range = self.plotItem.viewRange()
                    x_min, x_max = view_range[0]  # Get current x-axis (time) range
                    start_idx = bisect_left(x_data, x_min)
                    end_idx = bisect_right(x_data, x_max)
                    visible_y_data = y_data[start_idx:end_idx]

                    if len(visible_y_data) > 0:
                        value = stat_info['method'](visible_y_data)
                        formatted_value = format_str.format(value)
                        stat_info['label'].setText(formatted_value)
                    else:
                        stat_info['label'].setText(format_str.format(0))

    def add_vertical_line(self, channel, x):
        if channel in self.lines:
            self.removeItem(self.lines[channel])

        if x is None:
            return

        line_color = self.get_line_style(channel)[0]
        pen = pg.mkPen(color=line_color)

        self.lines[channel] = pg.InfiniteLine(pos=x, angle=90, movable=False, pen=pen)
        self.addItem(self.lines[channel])

    # Clear all line data.
    def reset(self):
        for line in self.lines.values():
            if isinstance(line, pg.PlotDataItem):
                line.setData([], [])
            elif isinstance(line, pg.InfiniteLine):
                self.removeItem(line)

        for stats in self.statistics.values():
            for format_str, stat_info in stats.items():
                stat_info['label'].setText(format_str.format(0))

    # Activates whenever the mouse hovers over or leaves the graph
    # - Makes the export button visible only when the mouse is hovering over the graph
    # - Makes the mouse coordinates appear when the mouse is hovering over the graph
    def event(self, event):
        if not self.full_init:
            return super().event(event)

        if event.type() == QEvent.HoverEnter:
            self.export_button.show()
        elif event.type() == QEvent.HoverLeave:
            self.mouse_label.setText("")
            self.export_button.hide()
        elif event.type() == QEvent.HoverMove:
            mouse_point = self.getPlotItem().getViewBox().mapSceneToView(event.position())
            view_range = self.getPlotItem().getViewBox().viewRange()

            if (view_range[0][0] <= mouse_point.x() <= view_range[0][1] and
                view_range[1][0] <= mouse_point.y() <= view_range[1][1]):
                self.mouse_label.setText(f"{mouse_point.x():.2f}, {mouse_point.y():.2f}")
            else:
                self.mouse_label.setText("")

        return super().event(event)

    def export_clicked(self):
        self.edit_dialog = QDialog(self)

        png_button = QPushButton("Export as PNG")
        csv_button = QPushButton("Export as CSV")
        png_button.clicked.connect(lambda x: self.export('png'))
        csv_button.clicked.connect(lambda x: self.export('csv'))
        png_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        csv_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # Ensure no button is selected by default
        png_button.setAutoDefault(False)
        csv_button.setAutoDefault(False)

        layout = QVBoxLayout()
        layout.addWidget(png_button)
        layout.addWidget(csv_button)

        self.edit_dialog.setWindowTitle("Export Graph")
        self.edit_dialog.setFixedSize(300, 150)
        self.edit_dialog.setLayout(layout)
        self.edit_dialog.show()

    def export(self, extension):
        self.edit_dialog.close()

        options = QFileDialog.Options()
        default_dir = QStandardPaths.writableLocation(QStandardPaths.DownloadLocation)
        filename = self.getPlotItem().titleLabel.text.lower().replace(' ', '_') + f"_plot.{extension}"
        filename = os.path.join(default_dir, filename)

        # Check if the file already exists and modify the filename accordingly
        counter = 1
        while os.path.exists(filename):
            filename = filename.replace('.csv', f" ({counter}).{extension}")
            counter += 1

        filename, _ = QFileDialog.getSaveFileName(
            self,
            f"Save {extension.upper()}",
            filename,
            f"{extension.upper()} Files (*.{extension})",
            options=options
        )

        if not filename:
            self.edit_dialog.close()
            return

        if not filename.endswith(f'.{extension}'):
            filename += f'.{extension}'

        if extension == 'csv':
            exporter = CSVExporter(self.plotItem)
        elif extension == 'png':
            exporter = pg.exporters.ImageExporter(self.plotItem)
            exporter.parameters()['width'] = 650
        else:
            raise RuntimeError(f"Unsupported file extension '{extension}'")

        exporter.export(filename)
        self.edit_dialog.close()

    def resizeEvent(self, event):
        """Update the export button position based on the plot's size."""
        super().resizeEvent(event)

        if hasattr(self, 'export_button'):
            plot_height = self.height()

            button_y = plot_height - 24  # 85% from the top

            # Update button position and size
            self.export_button.setPos(0, button_y)

    def get_view_state(self):
        """
        Check if the plot is currently fully zoomed out.
        Returns: bool indicating if the view shows the full data range
        """
        current_range = self.viewRange()[0]
        view_limits = self.getViewBox().state['limits']['xLimits']
        if view_limits is None:
            return True

        low_diff = current_range[0] - view_limits[0]
        hi_diff = view_limits[1] - current_range[1]
        return low_diff + hi_diff < 1

    def get_line_style(self, channel):
        theme = self.config_manager.get_settings('theme')

        match channel:
            case Channel.TARGET:
                return ACCENT_COLORS[theme]['light_green'], Qt.SolidLine
            case Channel.DEPRE_LOW | Channel.PRE_LOW:
                return ACCENT_COLORS[theme]['magenta'], Qt.SolidLine
            case Channel.DEPRE_UP | Channel.PRE_UP:
                return ACCENT_COLORS[theme]['blue'], Qt.SolidLine
            case Channel.HI_PRE_ORIG | HistStat.O_PRESS:
                return ACCENT_COLORS[theme]['yellow'], Qt.SolidLine
            case Channel.HI_PRE_SAMPLE | HistStat.S_PRESS:
                return ACCENT_COLORS[theme]['yellow'], Qt.DashLine
            case Channel.DEPRE_VALVE | HistStat.DO_SLOPE | HistStat.DO_SWITCH | Event.DEPRESSURIZE:
                return ACCENT_COLORS[theme]['cyan'], Qt.SolidLine
            case HistStat.DS_SLOPE | HistStat.DS_SWITCH:
                return ACCENT_COLORS[theme]['cyan'], Qt.DashLine
            case Channel.PRE_VALVE | HistStat.PO_SLOPE | HistStat.PO_SWITCH | Event.PRESSURIZE:
                return ACCENT_COLORS[theme]['red'], Qt.SolidLine
            case HistStat.PS_SLOPE | HistStat.PS_SWITCH:
                return ACCENT_COLORS[theme]['red'], Qt.DashLine
            case _:
                raise ValueError(f"Unknown channel: {channel}")

    def get_text_color(self):
        theme = self.config_manager.get_settings('theme')
        return THEME_COLOR_VALUES[theme]['foreground']['base']

    def get_background_color(self):
        theme = self.config_manager.get_settings('theme')
        return THEME_COLOR_VALUES[theme]['background']['base']

    def update_export_icon(self):
        theme = self.config_manager.get_settings('theme')
        color_info = THEME_COLOR_VALUES[theme]["foreground"]
        icon_color = color(color_info, state="icon")
        icon = url(icon_color, "file_export")[4:-1]  # Removes 'url(' at the start and ')' at the end
        self.export_button.setImageFile(icon)



from time import time
import numpy as np
from enum import Enum


class Channel(Enum):
    TARGET = 0          # Analog CH0: target pressure 
    DEPRE_LOW = 1       # Analog CH1: depressurization valve lower sensor
    DEPRE_UP = 2        # Analog CH2: depressurization valve upper sensor
    PRE_LOW = 3         # Analog CH3: pressurization valve lower sensor
    PRE_UP = 4          # Analog CH4: pressurization valve upper sensor
    HI_PRE_ORIG = 5     # Analog CH5: high pressure transducer at the origin
    HI_PRE_SAMPLE = 6   # Analog CH6: high pressure transducer at the sample
    PUMP = 7            # Digital CH0: high pressure pump (Active low / pumping on False)
    DEPRE_VALVE = 8     # Digital CH1: depressurize valve (Active low / open on False)
    PRE_VALVE = 9       # Digital CH2: pressurize valve (Active low / open on False)
    LOG = 10            # Digital CH4: log (Active low / logging on False)

    def __str__(self):
        match self.value:
            case 0:
                return "Target Pressure"
            case 1:
                return "Depressurization Valve Lower Sensor"
            case 2:
                return "Depressurization Valve Upper Sensor"
            case 3:
                return "Pressurization Valve Lower Sensor"
            case 4:
                return "Pressurization Valve Upper Sensor"
            case 5:
                return "Origin High Pressure Sensor"
            case 6:
                return "Sample High Pressure Sensor"
            case 7:
                return "Pump DIO"
            case 8:
                return "Depressurization Valve DIO"
            case 9:
                return "Pressurization Valve DIO"
            case 10:
                return "Log DIO"
            case _:
                raise ValueError(f"Unknown value: {self.value}")


# Enum for history statistics
class HistStat(Enum):
    O_PRESS = "Origin Pressure"
    S_PRESS = "Sample Pressure"
    DO_SLOPE = "Origin Depressurization Slope"
    DS_SLOPE = "Sample Depressurization Slope"
    PO_SLOPE = "Origin Pressurization Slope"
    PS_SLOPE = "Sample Pressurization Slope"
    DO_SWITCH = "Origin Depressurization Switch Time"
    DS_SWITCH = "Sample Depressurization Switch Time"
    PO_SWITCH = "Origin Pressurization Switch Time"
    PS_SWITCH = "Sample Pressurization Switch Time"

    def __str__(self):
        return self.value


# Returns a view of the selected channel
# Can be used on data that has not been wrapped in an event
def get_channel(data, channel):
    if isinstance(data, Event):
        data = data.data
    if channel.value <= 6:
        return data[:,channel.value]
    else:
        digital = 7
        bit = 1 << (channel.value - 7)
        if data[:,digital].dtype != np.int16:
            data = data.astype(np.int16)
        channel_data = data[:,digital] & bit
        return channel_data.astype(np.bool_)


# Used for ramp detection
def gaussian_filter(data, kernel_size, sigma=1):
    """Applies a 1D Gaussian filter to the data."""
    if kernel_size % 2 == 0 or kernel_size <= 0:
        raise ValueError("kernel_size must be a positive odd integer")

    kernel_range = np.arange(-kernel_size // 2 + 1, kernel_size // 2 + 1)
    kernel = np.exp(-kernel_range**2 / (2 * sigma**2))
    kernel /= kernel.sum()
    padded_data = np.pad(data, pad_width=kernel_size//2, mode='edge')
    return np.convolve(padded_data, kernel, mode='valid')


# Represents an event
# All methods which extract information for plotting are in this class
class Event:
    PRESSURIZE = 0
    DEPRESSURIZE = 1
    PERIOD = 2
    PRESSURE = 3
    PUMP = 4

    def __init__(self, event_type, data, event_index = None, event_time=None, step_time = None) -> None:
        if type(event_type) == int and 4 >= event_type >= 0:
            self.event_type = event_type
        else:
            raise RuntimeError(event_type + "event not supported.")

        if event_time is None:
            self.event_time = time()
            if data is not None:
                self.event_time -= data.shape[1] / 4000 # Assumes srate = 4000. 
            if event_index is not None:
                self.event_time += event_index / 4000
        # Case where this is a priorly generated event being deserialized
        else:
            self.event_time = event_time

        # Required for some math
        self.event_index = event_index # index where the actual event occured uint8
        if step_time is None and event_index is not None:
            self.step_time = 1 / 4 # Assumes srate = 4000. step_time is in ms.
        else:
            self.step_time = step_time

        # Period events can be long. Therefore, only take 600 data points to log and plot. (same as pressurize and depressurize plots)
        # No statistical analysis of period events is necessary, so this loss of data is fine.
        # If initialized with a step time, this means the data has already been compressed.
        if self.event_type == self.PERIOD or self.event_type == self.PUMP and step_time is None:
            self.data = self.compress_data(data, 600)
        else:
            self.data = data # np.ndarray (?,8) np.int16

    # used to call all info functions
    def get_event_info(self, hist_stat):
        match hist_stat:
            case HistStat.O_PRESS:
                return self.get_initial_origin()
            case HistStat.S_PRESS:
                return self.get_initial_sample()
            case HistStat.DO_SLOPE | HistStat.PO_SLOPE:
                return self.get_slope(Channel.HI_PRE_ORIG)
            case HistStat.DS_SLOPE | HistStat.PS_SLOPE:
                return self.get_slope(Channel.HI_PRE_SAMPLE)
            case HistStat.DO_SWITCH | HistStat.PO_SWITCH:
                return self.get_origin_switch_time()
            case HistStat.DS_SWITCH | HistStat.PS_SWITCH:
                return self.get_sample_switch_time()

    # Decrease size of data. Takes every nth data point, making sure not to lose valve events.
    def compress_data(self, data, num_points):
        if len(data) <= num_points:
            return data
        else:
            step = len(data) / num_points
            self.step_time = self.step_time * step
            indices = np.round(np.arange(0, num_points) * step).astype(int)

            compressed_data = data[indices]

            # Change where event occurred
            self.event_index = np.round(self.event_index / step).astype(int)

            # Recover lost valve events so they are plotted
            depressurize_indices = np.where(get_channel(data, Channel.DEPRE_VALVE) == np.min(get_channel(data, Channel.DEPRE_VALVE)))[0]
            pressurize_indices = np.where(get_channel(data, Channel.PRE_VALVE) == np.min(get_channel(data, Channel.PRE_VALVE)))[0]
            compressed_depressurize_indices = np.unique((depressurize_indices / step).astype(int))
            compressed_pressurize_indices = np.unique((pressurize_indices / step).astype(int))
            # Set to false with bitwise operation
            depre_mask = np.uint16(0b1111111111111101)
            pre_mask = np.uint16(0b1111111111111011)
            compressed_data[compressed_depressurize_indices,7] = compressed_data[compressed_depressurize_indices,7] & depre_mask
            compressed_data[compressed_pressurize_indices,7] = compressed_data[compressed_pressurize_indices,7] & pre_mask
            return compressed_data

    # Average of entire event
    def get_origin_pressure(self):
        if self.event_type != self.PRESSURE:
            raise RuntimeError("Cannot call get_origin_pressure() on event type " + self.event_type) 

        sample_pressure = np.average(get_channel(self, Channel.HI_PRE_ORIG))
        return sample_pressure

    # Average of entire event
    def get_target_pressure(self):
        if self.event_type != self.PRESSURE:
            raise RuntimeError("Cannot call get_target_pressure() on event type " + self.event_type) 

        target_pressure = np.average(get_channel(self, Channel.TARGET))
        return target_pressure

    # Returns pressure before event starts
    def get_initial_sample(self):
        if self.event_type != self.DEPRESSURIZE and self.event_type != self.PRESSURE:
            raise RuntimeError("Cannot call get_initial_sample() on event type " + self.event_type)

        data = get_channel(self, Channel.HI_PRE_SAMPLE)
        if self.event_type == self.DEPRESSURIZE:
            data = data[:self.event_index]

        return np.average(data)

    # Returns pressure before event starts
    def get_initial_origin(self):
        if self.event_type != self.DEPRESSURIZE and self.event_type != self.PRESSURE:
            raise RuntimeError("Cannot call get_initial_origin() on event type " + self.event_type)

        data = get_channel(self, Channel.HI_PRE_ORIG)
        if self.event_type == self.DEPRESSURIZE:
            data = data[:self.event_index]

        return np.average(data)

    # Returns pressure before event starts
    def get_initial_target(self):
        if self.event_type != self.PUMP:
            raise RuntimeError("Cannot call get_initial_target() on event type " + self.event_type)

        target_pressure = np.average(get_channel(self, Channel.TARGET)[:self.event_index])
        return target_pressure

    # Returns difference of index of half max and event_index
    def get_origin_switch_time(self):
        if self.event_type != self.DEPRESSURIZE and self.event_type != self.PRESSURIZE:
            raise RuntimeError("Cannot call get_origin_slope() on event type " + self.event_type)

        data = get_channel(self, Channel.HI_PRE_ORIG)
        minimum = np.min(data)
        data_range = np.max(data) - minimum
        half_max = minimum + (data_range / 2)
        half_max_index = np.argmin(np.abs(data - half_max))

        return half_max_index - self.event_index

    # Returns difference of index of half max and event_index
    def get_sample_switch_time(self):
        if self.event_type != self.DEPRESSURIZE and self.event_type != self.PRESSURIZE:
            raise RuntimeError("Cannot call get_origin_slope() on event type " + self.event_type)

        data = get_channel(self, Channel.HI_PRE_SAMPLE)
        minimum = np.min(data)
        data_range = np.max(data) - minimum
        half_max = minimum + (data_range / 2)
        half_max_index = np.argmin(np.abs(data - half_max))

        return half_max_index - self.event_index

    # Returns max slope of pressurization or depressurization
    def get_slope(self, channel):
        if self.event_type != self.DEPRESSURIZE and self.event_type != self.PRESSURIZE:
            raise RuntimeError(f"Cannot call get_slope() on event type {self.event_type}")

        y = get_channel(self, channel)

        # Calculate the first derivative of the data
        dy = np.diff(y)
        dy_smoothed = gaussian_filter(dy, 5, 5)

        if self.event_type == self.PRESSURIZE:
            slope = max(dy_smoothed)
        elif self.event_type == self.DEPRESSURIZE:
            slope = min(dy_smoothed)

        return slope



from PySide6.QtWidgets import (
    QGridLayout,
    QGroupBox,
    QLabel,
    QSizePolicy
)
from icarus_v2.backend.event import Event
from PySide6.QtCore import QCoreApplication
import numpy as np
from icarus_v2.backend.configuration_manager import ConfigurationManager


# Panel for counts of valves and such
class CounterDisplay(QGroupBox):
    def __init__(self):
        super().__init__()

        self.config_manager = ConfigurationManager()
        self.config_manager.settings_updated.connect(self.update_settings)
        self.counts = self.config_manager.get_settings("counter_settings")
        QCoreApplication.instance().aboutToQuit.connect(self.save_settings)

        self.pump_times = []

        # Counters
        self.pump_counter = QLabel(str(self.counts["pump_count"]))
        self.pressurize_counter = QLabel(str(self.counts["pressurize_count"]))
        self.depressurize_counter = QLabel(str(self.counts["depressurize_count"]))
        self.stroke_display = QLabel("0.00")

        # Labels
        pump_count_label = QLabel("Pump Count:")
        pressurize_count_label = QLabel("Pressurize Count:")
        depressurize_count_label = QLabel("Depressurize Count:")
        stroke_display_label = QLabel("Pump Strokes/hr:")

        # Set main layout
        layout = QGridLayout()
        layout.setSpacing(0)
        layout.addWidget(pump_count_label, 0, 0)
        layout.addWidget(pressurize_count_label, 1, 0)
        layout.addWidget(depressurize_count_label, 2, 0)
        layout.addWidget(stroke_display_label, 3, 0)
        layout.addWidget(self.pump_counter, 0, 1)
        layout.addWidget(self.pressurize_counter, 1, 1)
        layout.addWidget(self.depressurize_counter, 2, 1)
        layout.addWidget(self.stroke_display, 3, 1)

        self.setStyleSheet("font-size: 14pt;")
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        self.setLayout(layout)


    def increment_count(self, event):
        if event.event_type == Event.PUMP:
            self.counts["pump_count"] += 1
            self.pump_counter.setText(str(self.counts['pump_count']))

            # update pump strokes/hr
            self.pump_times.append(event.event_time)
            if len(self.pump_times) == 5:
                diff_in_sec = np.diff(self.pump_times).mean()
                strokes_per_hour = 3600 / diff_in_sec
                self.stroke_display.setText(f"{strokes_per_hour:.2f}")
                self.pump_times.pop(0)

        elif event.event_type == Event.PRESSURIZE:
            self.counts["pressurize_count"] += 1
            self.pressurize_counter.setText(str(self.counts['pressurize_count']))
        elif event.event_type == Event.DEPRESSURIZE:
            self.counts["depressurize_count"] += 1
            self.depressurize_counter.setText(str(self.counts['depressurize_count']))

        # Save counts to json every 10 updates
        if sum(self.counts.values()) % 10 == 0:
            # Do not emit so that this does not call self.update_settings
            self.save_settings()


    def save_settings(self):
        self.config_manager.save_settings("counter_settings", self.counts, emit=False)


    def update_settings(self, key):
        if key == 'counter_settings':
            self.counts = self.config_manager.get_settings(key)
            self.pump_counter.setText(str(self.counts['pump_count']))
            self.pressurize_counter.setText(str(self.counts['pressurize_count']))
            self.depressurize_counter.setText(str(self.counts['depressurize_count']))



from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import (
    QMainWindow,
    QVBoxLayout,
    QHBoxLayout,
    QGridLayout,
    QWidget,
    QGroupBox,
    QSizePolicy,
    QSpacerItem,
    QStackedWidget,
    QApplication,
    QLineEdit
)
from icarus_v2.gui.event_plot import EventPlot
from icarus_v2.gui.history_plot import HistoryPlot
from icarus_v2.gui.device_control_panel import DeviceControlPanel
from icarus_v2.gui.log_control_panel import LogControlPanel
from icarus_v2.gui.counter_display import CounterDisplay
from icarus_v2.gui.pressure_display import PressureDisplay
from icarus_v2.gui.tool_bar import ToolBar
from icarus_v2.gui.error_dialog import open_error_dialog
from icarus_v2.backend.log_reader import LogReader
from icarus_v2.backend.event import Event
from icarus_v2.backend.configuration_manager import ConfigurationManager
from icarus_v2.qdarktheme.load_style import load_stylesheet


# Can be either in log reading or device mode.
class MainWindow(QMainWindow):

    # Initializes all widgets and sets layout
    def __init__(self):
        super(MainWindow, self).__init__()
        self.mode = None
        self.data_handler = None
        self.config_manager = ConfigurationManager()
        self.update_theme()
        self.config_manager.settings_updated.connect(self.update_theme)

        # Window settings
        self.setWindowTitle("Icarus NMR")
        self.setMinimumSize(QSize(1000, 760))
        self.setStyleSheet("font-size: 17pt;")
        self.setAttribute(Qt.WA_AcceptTouchEvents, False)

        # Initialize all widgets
        self.pressure_event_display_range = (-10, 140)  # How much data to view around pressurize events
        self.pressurize_plot = EventPlot(Event.PRESSURIZE, parent=self)
        self.depressurize_plot = EventPlot(Event.DEPRESSURIZE, parent=self)
        self.period_plot = EventPlot(Event.PERIOD, parent=self)
        self.history_plot = HistoryPlot()
        self.device_control_panel = DeviceControlPanel()
        self.device_control_panel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.counter_display = CounterDisplay()
        self.pressure_display = PressureDisplay()

        # ToolBar
        self.toolbar = ToolBar(self)
        self.addToolBar(self.toolbar)
        self.toolbar.set_mode_signal.connect(self.set_mode)
        self.toolbar.history_reset_action.triggered.connect(self.reset_history)

        self.log_control_panel = LogControlPanel(tool_bar=self.toolbar)
        self.log_control_panel.pressurize_event_signal.connect(self.pressurize_plot.update_data)
        self.log_control_panel.pressurize_event_signal.connect(self.history_plot.render_event_time)
        self.log_control_panel.depressurize_event_signal.connect(self.depressurize_plot.update_data)
        self.log_control_panel.depressurize_event_signal.connect(self.history_plot.render_event_time)
        self.log_control_panel.period_event_signal.connect(self.period_plot.update_data)
        self.log_control_panel.event_list_signal.connect(self.history_plot.load_event_list)
        self.log_control_panel.reset_history_signal.connect(self.reset_history)
        self.log_control_panel.sample_sensor_connected.connect(self.set_sample_sensor_connected)
        self.log_control_panel.log_coefficients_signal.connect(self.pressurize_plot.set_log_coefficients)
        self.log_control_panel.log_coefficients_signal.connect(self.depressurize_plot.set_log_coefficients)
        self.log_control_panel.log_coefficients_signal.connect(self.period_plot.set_log_coefficients)
        self.log_control_panel.log_coefficients_signal.connect(self.history_plot.set_log_coefficients)
        self.log_control_panel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # Control and value layout
        # Show bounding box even when no controls are displayed
        self.bounding_box = QGroupBox()
        spacer = QSpacerItem(0, 0, QSizePolicy.Preferred, QSizePolicy.Expanding)
        box_layout = QVBoxLayout()
        box_layout.setSpacing(0)
        box_layout.addItem(spacer)
        self.bounding_box.setLayout(box_layout)
        self.bounding_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # Control layout
        control_layout = QGridLayout()
        control_layout.setSpacing(0)

        control_layout.addWidget(self.pressure_display, 0, 0)

        #Stacked Widget
        self.stacked_widget = QStackedWidget()
        self.stacked_widget.addWidget(self.device_control_panel)
        self.stacked_widget.addWidget(self.log_control_panel)
        self.stacked_widget.addWidget(self.bounding_box)
        self.stacked_widget.setCurrentWidget(self.bounding_box)
        self.stacked_widget.setFixedWidth(293)  # Width of pressure display
        self.stacked_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.stacked_widget.setContentsMargins(0, 0, 0, 0)
        control_layout.addWidget(self.stacked_widget)

        control_layout.addWidget(self.counter_display, 2, 0)
        control_layout.setContentsMargins(0, 0, 0, 0)
        control_layout.setColumnStretch(0, 1)

        # Event plot layout
        event_layout = QVBoxLayout()
        event_layout.setSpacing(0)
        event_layout.addWidget(self.pressurize_plot)
        event_layout.addWidget(self.depressurize_plot)
        event_layout.addWidget(self.period_plot)

        # Set layout
        self.main_layout = QHBoxLayout()
        self.main_layout.addLayout(event_layout)
        self.main_layout.addWidget(self.history_plot)
        self.main_layout.addLayout(control_layout)

        # Add layout to dummy widget and apply to main window
        widget = QWidget()
        widget.setLayout(self.main_layout)
        self.setCentralWidget(widget)

        self.connected = False
        self.set_mode("device")

        # disable focus for all widgets except QTextEdit
        for widget in self.findChildren(QWidget):
            if not isinstance(widget, QLineEdit):
                widget.setFocusPolicy(Qt.NoFocus)

    def update_theme(self, settings_key: str | None = None) -> None:
        # skip setting theme if this is triggered by settings updates
        if settings_key is not None and settings_key != "theme":
            return

        theme = self.config_manager.get_settings('theme')
        app = QApplication.instance()

        if app is not None:
            app.setStyleSheet(load_stylesheet(theme))

    # Connects widgets to backend
    def set_device(self, data_handler):
        self.data_handler = data_handler
        self.device_control_panel.set_pulse_generator(data_handler.pulse_generator)
        self.log_control_panel.log_reader.set_logger(data_handler.logger)
        self.toolbar.set_pressure_signal(data_handler.pressure_event_signal)
        self.toolbar.set_sentry(data_handler.sentry)

        # Connections that are never disconnected
        data_handler.pressure_event_signal.connect(self.pressure_display.update_pressure)
        data_handler.pressurize_event_signal.connect(self.counter_display.increment_count)
        data_handler.depressurize_event_signal.connect(self.counter_display.increment_count)
        data_handler.pump_event_signal.connect(self.counter_display.increment_count)
        data_handler.acquiring_signal.connect(lambda x: self.set_connected(x))
        data_handler.toolbar_warning.connect(self.toolbar.display_warning)
        # Ideally this is unnecessary as the signal should be directly sent to the pulse_generator. This is a workaround
        data_handler.shutdown_signal.connect(self.device_control_panel.on_shutdown)

        def error_dialog(err):
            self.dialog = open_error_dialog(err)
        data_handler.display_error.connect(error_dialog)

        # Connect plot signals
        self.set_mode(self.mode)

    # Set device signal connections, clear history, set control panel
    def set_mode(self, mode):
        self.mode = mode
        if mode == "log":
            
            self.stacked_widget.setCurrentWidget(self.log_control_panel)

            # Disconnect device signals from gui elements (excludes pressure display and counter display)
            if self.data_handler is not None:
                self.data_handler.pressurize_event_signal.disconnect(self.pressurize_plot.update_data)
                self.data_handler.depressurize_event_signal.disconnect(self.depressurize_plot.update_data)
                self.data_handler.period_event_signal.disconnect(self.period_plot.update_data)
                self.data_handler.pressurize_event_signal.disconnect(self.history_plot.add_event)
                self.data_handler.depressurize_event_signal.disconnect(self.history_plot.add_event)
                self.data_handler.pressure_event_signal.disconnect(self.history_plot.add_event)
                self.data_handler.log_signal.disconnect(self.reset_history)
                self.data_handler.sample_sensor_connected.disconnect(self.set_sample_sensor_connected)

            # Clear plots
            self.reset_history()

            # If no device connected, automatically open log fileDialog
            if not self.connected:
                self.log_control_panel.choose_log()

        elif mode == "device":
            if self.connected:
                self.stacked_widget.setCurrentWidget(self.device_control_panel)
            else:
                self.stacked_widget.setCurrentWidget(self.bounding_box)

            # Connect device event signals to GUI elements
            if self.data_handler is not None:
                self.data_handler.pressurize_event_signal.connect(self.pressurize_plot.update_data)
                self.data_handler.depressurize_event_signal.connect(self.depressurize_plot.update_data)
                self.data_handler.period_event_signal.connect(self.period_plot.update_data)
                self.data_handler.pressurize_event_signal.connect(self.history_plot.add_event)
                self.data_handler.depressurize_event_signal.connect(self.history_plot.add_event)
                self.data_handler.pressure_event_signal.connect(self.history_plot.add_event)
                self.data_handler.log_signal.connect(self.reset_history)
                self.data_handler.sample_sensor_connected.connect(self.set_sample_sensor_connected)

            # Clear plots
            self.reset_history()
            self.log_control_panel.reset()

            # restore history
            if self.connected:
                current_log_file = self.data_handler.logger.filename
                self.data_handler.logger.flush()
                reader = LogReader()
                reader.read_events(current_log_file)
                self.history_plot.load_event_list(reader.events)
                press = None
                depress = None
                per = None
                for event in reader.events[::-1]:
                    if press is None and event.event_type == Event.PRESSURIZE:
                        press = event
                    elif depress is None and event.event_type == Event.DEPRESSURIZE:
                        depress = event
                    elif per is None and event.event_type == Event.PERIOD:
                        per = event
                    if press is not None and depress is not None and per is not None:
                        break
                self.pressurize_plot.update_data(press)
                self.depressurize_plot.update_data(depress)
                self.period_plot.update_data(per)

    # Set control panel, clear pressure
    def set_connected(self, connected):
        self.connected = connected
        self.toolbar.set_connected(connected)
        self.log_control_panel.set_logging(connected)
        if connected:
            self.device_control_panel.reset()
            if self.mode == "device":
                self.stacked_widget.setCurrentWidget(self.device_control_panel)
        else:
            self.pressure_display.reset()
            if self.mode == "device":
                self.stacked_widget.setCurrentWidget(self.bounding_box)

    def set_sample_sensor_connected(self, connected):
        self.history_plot.set_sample_sensor(connected)
        self.pressurize_plot.set_sample_sensor(connected)
        self.depressurize_plot.set_sample_sensor(connected)
        self.period_plot.set_sample_sensor(connected)

    def reset_history(self):
        self.set_sample_sensor_connected(True)

        self.history_plot.reset_history()
        self.pressurize_plot.plot.reset()
        self.depressurize_plot.plot.reset()
        self.period_plot.plot.reset()

    # Runs on quitting the application
    def closeEvent(self, event):
        super().closeEvent(event)

        if self.data_handler is not None:
            self.data_handler.quit()

        if self.toolbar.settings_dialog is not None:
            self.toolbar.settings_dialog.close()



from PySide6.QtCore import QThread, Signal
from icarus_v2.backend.ring_buffer import RingBuffer, SPMCRingBufferReader
import numpy as np
import usb.core
from icarus_v2.backend.logger import Logger


# This class is responsible for creating a buffer, reading from the device, processing the data, and putting it into the buffer.
class BufferLoader(QThread):
    device_disconnected = Signal()

    def __init__(self, buffer_seconds=120) -> None:
        super().__init__()
        num_channels = 8
        buffer_capacity = int(buffer_seconds * 4000)
        self.buffer = RingBuffer((buffer_capacity, num_channels), np.int16)
        self.device = None

        # TESTING ONLY. logs all example data.
        self.log_raw = False


    def set_device(self, device):
        self.device = device


    def run(self):
        if self.device is None: return

        if self.log_raw:
            self.raw_logger = Logger(is_raw=True)
            self.raw_logger.new_log_file()

        self.device.start_scan()

        while self.device.acquiring:
            try:
                data = self.device.read_data()
            except usb.core.USBError:
                # Device disconnected
                self.device.acquiring = False
                self.device_disconnected.emit()
                return
            except RuntimeError as e:
                if "End of file reached." in str(e):
                    # End of file reached
                    self.device.acquiring = False
                    self.device_disconnected.emit()
                    return
                else:
                    raise e

            if self.log_raw:
                self.raw_logger.log_raw(data)
            processed_data = self.process_data(data)
            self.buffer.enqueue(processed_data)

        self.device.end_scan()


    def process_data(self, data):
        data_shape = (self.device.points_to_read, self.device.channels_to_read)

        int_array = np.frombuffer(data, dtype=np.int16)
        reshaped_array = np.reshape(int_array, data_shape)
        reshaped_array[:,-1] = reshaped_array[:,-1] >> 8    # Digital is only in the 1st byte

        return reshaped_array


    def new_reader(self):
        return SPMCRingBufferReader(self.buffer)


    def get_sample_rate(self):
        return self.device.sample_rate


    def quit(self):
        self.may_start = False
        if self.device is not None and self.device.acquiring:
            self.device.stop()
            self.device.close_device()
